"""
class TestDocTrackingPandasDataframe:
    def test_doc(self):
        # Python 3.6.8
        from dbnd import log_metric, log_dataframe
        import pandas as pd

        transactions_df = pd.read_csv("~/data/example.csv")

        # log dataframe
        log_dataframe(
            "Full table", transactions_df, with_histograms=False, with_preview=False
        )

        # log_metrics
        minimum_amount = 5000
        num_large_transactions = transactions_df[
            transactions_df["transaction_amt"] >= minimum_amount
        ].shape[0]
        avg_large_transaction = transactions_df[
            transactions_df["transaction_amt"] >= minimum_amount
        ].mean()["transaction_amt"]
        large_transactions_df = transactions_df[
            transactions_df["transaction_amt"] >= minimum_amount
        ]

        log_metric("Number of large transactions(>= 5000)", num_large_transactions)
        log_metric("mean large transactions", avg_large_transaction)
        log_dataframe(
            "Large transactions (>= 5000)", large_transactions_df, with_preview=False
        )



    def test_process_customers_data(self):
        #### DOC START
        @task
        def prepare_data(data: pd.DataFrame = "s3://pipelines/customers_data.csv") -> pd.DataFrame:
            log_dataframe("customers_data", data,
                          with_histograms=LogDataRequest(include_all_numeric=True,
                                                         exclude_columns=["name", "phone"]))
        #### DOC END
        prepare_data.dbnd_run()
"""



import datetime
import itertools
import logging

from typing import List, Tuple

import matplotlib
import numpy as np
import pandas as pd

from matplotlib import figure
from sklearn import preprocessing
from sklearn.linear_model import ElasticNet
from sklearn.metrics import mean_absolute_error, mean_squared_error, r2_score
from sklearn.model_selection import train_test_split

from dbnd import log_dataframe, log_metric, parameter, pipeline, task
from dbnd.utils import data_combine, period_dates
from dbnd_examples.data import demo_data_repo
from dbnd_examples.orchestration.examples.wine_quality import package_as_docker
from targets import target


logger = logging.getLogger(__name__)
matplotlib.use("Agg")


def data_source(task_target_date, name):
    return target(demo_data_repo.partner_to_file(name, task_target_date))


@task
def get_and_enrich(raw_data: pd.DataFrame, column_name: str) -> pd.DataFrame:
    scaler = preprocessing.MinMaxScaler()
    raw_data[column_name + "_norm"] = scaler.fit_transform(
        raw_data[[column_name]].values.astype(float)
    )
    return raw_data


@task
def clean_data(raw_data: pd.DataFrame) -> pd.DataFrame:
    return raw_data.fillna(0)


@pipeline
def ingest_partner_a(task_target_date):
    raw_data = data_source(name="a", task_target_date=task_target_date)
    clean = clean_data(raw_data=raw_data)
    return get_and_enrich(raw_data=clean, column_name="0")


@pipeline
def ingest_partner_b(task_target_date):
    raw_data = data_source(name="b", task_target_date=task_target_date)
    data = clean_data(raw_data=raw_data)
    return data


@pipeline
def ingest_partner_c(task_target_date):
    raw_data = data_source(name="c", task_target_date=task_target_date)
    clean = clean_data(raw_data=raw_data)
    data = get_and_enrich(raw_data=clean, column_name="10")
    return data


@pipeline
def fetch_partner_data(
    task_target_date, selected_partners: List[str], period=datetime.timedelta(days=7)
) -> List[pd.DataFrame]:
    partner_data = []
    for partner in selected_partners:
        all_data = []
        for d in period_dates(task_target_date, period):
            if partner == "a":
                data = ingest_partner_a(task_target_date=d)
            elif partner == "b":
                data = ingest_partner_b(task_target_date=d)
            elif partner == "c":
                data = ingest_partner_c(task_target_date=d)
            else:
                raise Exception("Partner not found!")

            all_data.append(data)
        partner_data.append(data_combine(all_data, sort=True))
    return partner_data


@task
def calculate_features(
    raw_data: List[pd.DataFrame], selected_features: List[str] = None
) -> pd.DataFrame:
    result = raw_data.pop(0)
    for d in raw_data:
        result = result.merge(d, on="id")
    if selected_features:
        result = result[selected_features]
    return result


@task(result="training_set, test_set, validation_set")
def split_data(
    raw_data: pd.DataFrame,
) -> Tuple[pd.DataFrame, pd.DataFrame, pd.DataFrame]:
    columns_to_remove = set(["id", "0_norm", "10_norm"])
    if columns_to_remove.issubset(raw_data.columns):
        raw_data.drop(columns_to_remove, axis=1, inplace=True)

    train_df, test_df = train_test_split(raw_data)
    test_df, validation_df = train_test_split(test_df, test_size=0.5)

    log_dataframe("raw", raw_data)
    log_metric("target.mean", raw_data["target"].mean())
    log_metric("target.std", raw_data["target"].std())

    return train_df, test_df, validation_df


@task
def train_model(
    test_set: pd.DataFrame,
    training_set: pd.DataFrame,
    alpha: float = 1.0,
    l1_ratio: float = 0.5,
) -> ElasticNet:
    lr = ElasticNet(alpha=alpha, l1_ratio=l1_ratio)
    lr.fit(training_set.drop(["target"], 1), training_set[["target"]])
    prediction = lr.predict(test_set.drop(["target"], 1))

    (rmse, mae, r2) = calculate_metrics(test_set[["target"]], prediction)

    log_metric("rmse", rmse)
    log_metric("mae", mae)
    log_metric("r2", r2)

    logging.info(
        "Elasticnet model (alpha=%f, l1_ratio=%f): rmse = %f, mae = %f, r2 = %f",
        alpha,
        l1_ratio,
        rmse,
        mae,
        r2,
    )
    return lr


def _create_scatter_plot(actual, predicted) -> figure.Figure:
    import matplotlib.pyplot as plt

    fig = plt.figure()
    ax = fig.add_subplot(1, 1, 1)
    ax.set_title("Actual vs. Predicted")
    ax.set_xlabel("Actual Labels")
    ax.set_ylabel("Predicted Values")
    ax.scatter(actual, predicted)
    return fig


@task(result=("report", "prediction_scatter_plot"))
def validate_model_for_customer(
    model: ElasticNet, validation_dataset: pd.DataFrame, threshold=0.2
) -> Tuple[str, figure.Figure]:
    log_dataframe("validation", validation_dataset)
    # support for py3 parqeut
    validation_dataset = validation_dataset.rename(str, axis="columns")
    validation_x = validation_dataset.drop(["target"], 1)
    validation_y = validation_dataset[["target"]]

    prediction = model.predict(validation_x)
    (rmse, mae, r2) = calculate_metrics(validation_y, prediction)

    log_metric("rmse", rmse)
    log_metric("mae", mae)
    log_metric("r2", r2)
    fig = _create_scatter_plot(validation_y, prediction)
    if r2 < threshold:
        raise Exception(
            "Model quality is below threshold. Got R2 equal to %s, expect at least %s"
            % (r2, threshold)
        )

    return "%s,%s,%s" % (rmse, mae, r2), fig


@pipeline(result=("model", "validation"))
def train_model_for_customer(
    task_target_date,
    data: pd.DataFrame = None,
    alpha: float = 1.0,
    l1_ratio: float = 0.5,
    period=datetime.timedelta(days=7),
    selected_features: List[str] = None,
):
    if data is None:
        partners = fetch_partner_data(task_target_date=task_target_date, period=period)
        data = calculate_features(
            selected_features=selected_features, raw_data=partners
        )
    training_set, test_set, validation_set = split_data(raw_data=data)

    model = train_model(
        test_set=test_set, training_set=training_set, alpha=alpha, l1_ratio=l1_ratio
    )
    validation, _ = validate_model_for_customer(
        model=model, validation_dataset=validation_set
    )
    return model, validation


@pipeline
def train_for_all_customer(customer=parameter[List[pd.DataFrame]]):
    result = {}
    for c in customer:
        (model, validation) = train_model_for_customer(data=c)
        result[str(c)] = (model, validation)
    return result


@pipeline
def training_with_parameter_search(
    data: pd.DataFrame = demo_data_repo.sample_customer_file,
    alpha_step: float = 0.3,
    l1_ratio_step: float = 0.4,
):
    result = {}
    variants = list(
        itertools.product(np.arange(0, 1, alpha_step), np.arange(0, 1, l1_ratio_step))
    )
    logger.info("All Variants: %s", variants)
    for alpha_value, l1_ratio in variants:
        exp_name = "Predict_%f_l1_ratio_%f" % (alpha_value, l1_ratio)
        model, validation = train_model_for_customer(
            data=data, alpha=alpha_value, l1_ratio=l1_ratio, task_name=exp_name
        )

        result[exp_name] = (model, validation)
    return result


@pipeline(result=("model", "validation", "serving"))
def tran_model_and_package():
    model, validation = train_model_for_customer()
    serving = package_as_docker(model=model)
    return model, validation, serving


# Uttilities
def calculate_metrics(actual, pred):
    rmse = np.sqrt(mean_squared_error(actual, pred))
    mae = mean_absolute_error(actual, pred)
    r2 = r2_score(actual, pred)
    return rmse, mae, r2



import sys

from dbnd import dbnd_tracking_start, task


@task
def some_function(x="my_default"):
    return x[0]


def my_f():
    some_function()
    some_function("my_value")


if __name__ == "__main__":
    dbnd_tracking_start()
    print("Main script is running")
    print("Command lines are: {} ", sys.argv)
    with open(sys.argv[1], "w") as fp:
        fp.write("Success")
    my_f()
    sys.exit(1)



from dbnd.testing.helpers_pytest import assert_run_task
from dbnd_examples.orchestration.examples.report import dump_db, partners_report


class TestTutorialReportExamples(object):
    def test_tutorial_report_dump(self):
        task = assert_run_task(dump_db.task("aa,bb"))
        assert task

    def test_tutorial_partners_report(self):
        task = assert_run_task(partners_report.task([1, 2]))
        assert task



import logging
import sys

from dbnd import output, relative_path, task
from dbnd.tasks.basics.shell import bash_cmd
from targets.types import PathStr


logger = logging.getLogger(__name__)


def cli_scripts(*path):
    return relative_path(__file__, *path)


@task(my_cli_output=output[PathStr])
def my_cli(parameter, my_cli_input, my_cli_output):
    # type: (int, PathStr, PathStr) -> None
    bash_cmd(
        args=[
            sys.executable,
            cli_scripts("../../tracking/tracking_script.py"),
            my_cli_output,
            parameter,
            my_cli_input,
        ],
        check_retcode=1,
    )



from dbnd_airflow.scheduler.dags_provider_from_file import get_dags_from_file


# airflow will only scan files containing the text DAG or airflow. This comment performs this function

dags = get_dags_from_file()
if dags:
    globals().update(dags)



"""
Code that goes along with the Airflow tutorial located at:
https://github.com/airbnb/airflow/blob/master/airflow/example_dags/tutorial.py
"""
from datetime import datetime, timedelta

from airflow import DAG
from airflow.operators.bash_operator import BashOperator


default_args = {
    "owner": "airflow",
    "depends_on_past": False,
    "start_date": datetime.now() - timedelta(days=7),
    "email": ["airflow@example.com"],
    "email_on_failure": False,
    "email_on_retry": False,
    "retries": 1,
    "retry_delay": timedelta(minutes=5),
    # 'queue': 'bash_queue',
    # 'pool': 'backfill',
    # 'priority_weight': 10,
    # 'end_date': datetime(2016, 1, 1),
}

with DAG("bash_dag", default_args=default_args) as dag:
    # t1, t2 and t3 are examples of tasks created by instantiating operators
    t1 = BashOperator(task_id="print_date", bash_command="date", dag=dag)
    t2 = BashOperator(task_id="sleep", bash_command="sleep 1", retries=3, dag=dag)

    templated_command = """
        {% for i in range(5) %}
            echo "{{ ds }}"
            echo "{{ macros.ds_add(ds, 7)}}"
            echo "{{ params.my_param }}"
        {% endfor %}
    """
    t3 = BashOperator(
        task_id="templated",
        bash_command=templated_command,
        params={"my_param": "Parameter I passed in"},
        dag=dag,
    )

    t2.set_upstream(t1)
    t3.set_upstream(t1)



from dbnd import log_metric


class TestDocTrackingQuickstart:
    def test_step_one(self):
        #### DOC START
        # Python 3.6.8
        from typing import Tuple

        from pandas import DataFrame, Series
        from sklearn import datasets
        from sklearn.model_selection import train_test_split

        def prepare_data() -> Tuple[DataFrame, DataFrame]:
            """load dataset from sklearn. split into training and testing sets"""
            raw_data = datasets.load_diabetes()

            # create a pandas DataFrame from sklearn dataset
            df = DataFrame(raw_data["data"], columns=raw_data["feature_names"])
            df["target"] = Series(raw_data["target"])

            # split the data into training and testing sets
            training_data, testing_data = train_test_split(df, test_size=0.25)

            return training_data, testing_data

        #### DOC END
        prepare_data()

    def test_integrating_dataframe_tracking(self):
        #### DOC START
        # Python 3.6.8
        import logging

        from typing import Tuple

        from pandas import DataFrame, Series
        from sklearn import datasets
        from sklearn.model_selection import train_test_split

        from dbnd import log_dataframe, log_metric

        logging.basicConfig(level=logging.INFO)

        def prepare_data() -> Tuple[DataFrame, DataFrame]:
            """load dataset from sklearn. split into training and testing sets"""
            raw_data = datasets.load_diabetes()

            # create a pandas DataFrame from sklearn dataset
            df = DataFrame(raw_data["data"], columns=raw_data["feature_names"])
            df["target"] = Series(raw_data["target"])

            # split the data into training and testing sets
            training_data, testing_data = train_test_split(df, test_size=0.25)

            # use DBND logging features to log DataFrames with histograms
            log_dataframe(
                "training data",
                training_data,
                with_histograms=True,
                with_schema=True,
                with_stats=True,
            )
            log_dataframe("testing_data", testing_data)

            # use DBND logging features to log the mean of s1
            log_metric("mean s1", training_data["s1"].mean())

            return training_data, testing_data

        #### DOC END
        prepare_data()

    def test_step_two(self):
        #### DOC START
        # Python 3.6.8
        import logging

        from typing import Tuple

        from pandas import DataFrame, Series
        from sklearn import datasets
        from sklearn.linear_model import LinearRegression
        from sklearn.metrics import mean_squared_error
        from sklearn.model_selection import train_test_split

        from dbnd import log_dataframe

        logging.basicConfig(level=logging.INFO)

        def prepare_data() -> Tuple[DataFrame, DataFrame]:
            """load dataset from sklearn. split into training and testing sets"""
            raw_data = datasets.load_diabetes()

            # create a pandas DataFrame from sklearn dataset
            df = DataFrame(raw_data["data"], columns=raw_data["feature_names"])
            df["target"] = Series(raw_data["target"])

            # split the data into training and testing sets
            training_data, testing_data = train_test_split(df, test_size=0.25)

            # use DBND logging features to log DataFrames with histograms
            log_dataframe(
                "training data",
                training_data,
                with_histograms=True,
                with_schema=True,
                with_stats=True,
            )
            log_dataframe("testing_data", testing_data)

            # use DBND logging features to log the mean of s1
            log_metric("mean s1", training_data["s1"].mean())

            return training_data, testing_data

        def train_model(training_data: DataFrame) -> LinearRegression:
            """train a linear regression model"""
            model = LinearRegression()

            # train a linear regression model
            model.fit(training_data.drop("target", axis=1), training_data["target"])
            return model

        def test_model(model: LinearRegression, testing_data: DataFrame) -> str:
            """test the model, output mean squared error and r2 score"""
            testing_x = testing_data.drop("target", axis=1)
            testing_y = testing_data["target"]
            predictions = model.predict(testing_x)

            mse = mean_squared_error(testing_y, predictions)
            r2_score = model.score(testing_x, testing_y)

            return f"MSE: {mse}, R2: {r2_score}"

        #### DOC END
        test_model(
            model=train_model(training_data=prepare_data()[0]),
            testing_data=prepare_data()[1],
        )

    def test_tracking_more_metrics(self):
        #### DOC START
        import logging

        from typing import Tuple

        from pandas import DataFrame, Series
        from sklearn import datasets
        from sklearn.linear_model import LinearRegression
        from sklearn.metrics import mean_squared_error
        from sklearn.model_selection import train_test_split

        from dbnd import log_dataframe, log_metric

        logging.basicConfig(level=logging.INFO)

        def prepare_data() -> Tuple[DataFrame, DataFrame]:
            """load dataset from sklearn. split into training and testing sets"""
            raw_data = datasets.load_diabetes()

            # create a pandas DataFrame from sklearn dataset
            df = DataFrame(raw_data["data"], columns=raw_data["feature_names"])
            df["target"] = Series(raw_data["target"])

            # split the data into training and testing sets
            training_data, testing_data = train_test_split(df, test_size=0.25)

            # use DBND logging features to log DataFrames with histograms
            log_dataframe(
                "training data",
                training_data,
                with_histograms=True,
                with_schema=True,
                with_stats=True,
            )
            log_dataframe("testing_data", testing_data)

            # use DBND logging features to log the mean of s1
            log_metric("mean s1", training_data["s1"].mean())

            return training_data, testing_data

        def train_model(training_data: DataFrame) -> LinearRegression:
            """train a linear regression model"""
            model = LinearRegression()

            # train a linear regression model
            model.fit(training_data.drop("target", axis=1), training_data["target"])

            # use DBND log crucial details about the regression model with log_metric:
            log_metric("model intercept", model.intercept_)  # logging a numeric
            log_metric("coefficients", model.coef_)  # logging an np array
            return model

        def test_model(model: LinearRegression, testing_data: DataFrame) -> str:
            """test the model, output mean squared error and r2 score"""
            testing_x = testing_data.drop("target", axis=1)
            testing_y = testing_data["target"]
            predictions = model.predict(testing_x)
            mse = mean_squared_error(testing_y, predictions)
            r2_score = model.score(testing_x, testing_y)

            # use DBND log_metric to capture important model details:
            log_metric("mean squared error:", mse)
            log_metric("r2 score", r2_score)

            return f"MSE: {mse}, R2: {r2_score}"

        # if __name__ == '__main__':
        training_set, testing_set = prepare_data()
        model = train_model(training_set)
        metrics = test_model(model, testing_set)
        #### DOC END

        assert metrics



'''
class TestDocTrackingSql:
    def test_doc(self):
        #### DOC START
        # Python 3.6.8
        from sqlalchemy import create_engine
        from dbnd import log_metric

        conn_string = 'postgresql+psycopg2://dbnd_postgres:postgres@localhost/dbnd_postgres'
        pg_db = create_engine(conn_string)

        minimum_amount = 5000

        # logging count of rows where transaction_amt >= 5000
        num_large_transactions_query = f"""SELECT COUNT(transaction_amt) FROM transactions_data WHERE transaction_amt >= {minimum_amount};"""
        results = pg_db.execute(num_large_transactions_query)
        num_large_transactions = results.fetchone()[0]
        log_metric("Number of large transactions(>= 5000)", num_large_transactions)

        # logging average of rowsw where transaction_amt >= 5000
        avg_large_transactions_query = f"""SELECT AVG(transaction_amt) FROM transactions_data WHERE transaction_amt >= {minimum_amount};"""
        results = pg_db.execute(avg_large_transactions_query)
        avg_large_transactions = results.fetchone()[0]
        log_metric("mean large transactions", avg_large_transactions)

        # logging all rows where transaction_amt >= 5000
        large_transactions_query = f"""SELECT * FROM transactions_data WHERE transaction_amt >= {minimum_amount};"""
        results = pg_db.execute(large_transactions_query)
        large_transactions = [row for row in results]
        log_metric("large transactions rows", large_transactions)

        # logging table shape
        num_rows_query = f"""SELECT COUNT (*) FROM transactions_data"""
        results = pg_db.execute(num_rows_query)
        num_rows = results.fetchone()[0]
        num_cols_query = f"""SELECT COUNT(column_name) FROM INFORMATION_SCHEMA.COLUMNS WHERE table_name = 'transactions_data';"""
        results = pg_db.execute(num_cols_query)
        num_cols = results.fetchone()[0]

        log_metric("table shape", (num_rows, num_cols))
        #### DOC END
'''



import datetime
import logging
import os

from random import randint, random

from mlflow import (
    active_run,
    end_run,
    get_tracking_uri,
    log_artifacts,
    log_metric,
    log_param,
    start_run,
)
from mlflow.tracking import MlflowClient

from dbnd import task


logger = logging.getLogger(__name__)


@task
def task_with_mflow(check_time=datetime.datetime.now()):
    # type: ( datetime.datetime)-> str
    logger.info("Running MLFlow tracking integration check!")
    logger.info("MLFlow tracking URI: {}".format(get_tracking_uri()))

    start_run()

    # params
    log_param("param1", randint(0, 100))
    log_param("param2", randint(0, 100))

    # metrics
    log_metric("foo1", random())
    log_metric("foo1", random() + 1)
    log_metric("foo2", random())
    log_metric("foo2", random() + 1)

    # artifacts
    if not os.path.exists("outputs"):
        os.makedirs("outputs")
    with open("outputs/test1.txt", "w") as f1, open("outputs/test2.txt", "w") as f2:
        f1.write("hello")
        f2.write("world!")
    log_artifacts("outputs")

    # Get run metadata & data from the tracking server
    service = MlflowClient()
    run_id = active_run().info.run_id
    run = service.get_run(run_id)
    logger.info("Metadata & data for run with UUID %s: %s" % (run_id, run))

    end_run()

    logger.info("MLFlow tracking integration check completed!")



import airflow

from airflow.contrib.hooks.snowflake_hook import SnowflakeHook
from airflow.models import DAG, BaseOperator

from dbnd_snowflake import snowflake_query_tracker
from dbnd_snowflake.airflow_operators import (
    LogSnowflakeResourceOperator,
    LogSnowflakeTableOperator,
)


args = {"start_date": airflow.utils.dates.days_ago(2), "owner": "databand"}

snowflake_conn_id = "test_snowflake_conn"
database = "SNOWFLAKE_SAMPLE_DATA"
schema = "TPCDS_SF100TCL"
table = "CUSTOMER"
select_query = f"select * from {database}.{schema}.{table} limit 1000"


dag_examples = [
    DAG(
        dag_id="snowflake_example_dag_{}".format(i),
        default_args=args,
        schedule_interval="0 1 * * *",  # Daily
    )
    for i in range(1, 4)
]
for dag in dag_examples:
    globals()[dag.dag_id] = dag


# EXAMPLE 1: Query Tracker as Context Manager
class Example1Operator(BaseOperator):
    def execute(self, context):
        hook = SnowflakeHook(snowflake_conn_id=snowflake_conn_id)
        with snowflake_query_tracker(database=database):
            hook.run(sql=select_query)


with dag_examples[0]:
    ex1_task = Example1Operator(task_id="example1")


# EXAMPLE 2: Query Tracker as Decorator -- not possible now, since connection string need to be passed
class Example2Operator(BaseOperator):
    def execute(self, context):
        hook = SnowflakeHook(snowflake_conn_id=snowflake_conn_id)
        with snowflake_query_tracker(database=database):
            hook.run(sql=select_query)


with dag_examples[1]:
    ex2_task = Example2Operator(task_id="example1")


# EXAMPLE 3: pass session_id, query_id over xcom to another operator
class Example3Operator(BaseOperator):
    def execute(self, context):
        hook = SnowflakeHook(snowflake_conn_id=snowflake_conn_id)
        with snowflake_query_tracker(database=database) as st:
            hook.run(select_query)
            session_id, query_id = st.get_last_session_with_query_id(many=False)

        context["ti"].xcom_push(key="session_id", value=session_id)
        context["ti"].xcom_push(key="query_id", value=query_id)


with dag_examples[2]:
    ex3_task = Example3Operator(task_id="example3")
    log_snowflake_table_task = LogSnowflakeTableOperator(
        table=table,
        snowflake_conn_id=snowflake_conn_id,
        warehouse=None,
        database=database,
        schema=schema,
        task_id="log_snowflake_table_task",
        key=f"snowflake_table_{table}",
    )

    log_snowflake_resources_task = LogSnowflakeResourceOperator(
        session_id="{{ti.xcom_pull(key='session_id') }}",
        query_id="{{ti.xcom_pull(key='query_id')[0] }}",
        snowflake_conn_id=snowflake_conn_id,
        warehouse=None,
        database=database,
        schema=schema,
        task_id="log_snowflake_resources_task",
        key="snowflake_query_resources",
    )
    ex3_task >> log_snowflake_table_task >> log_snowflake_resources_task



import logging

from dbnd.testing.helpers_pytest import assert_run_task
from dbnd_examples.orchestration.tutorial_syntax.T12_data_pandas_and_numpy import (
    f_test_pandas_numpy_flow,
)


logger = logging.getLogger(__name__)


class TestTutorialSyntax(object):
    def test_pandas_numpy(self):
        assert_run_task(f_test_pandas_numpy_flow.task())



class TestDocExplicitDatasetLogging:
    def test_doc(self):
        """
        #### DOC START
        from dbnd import dataset_op_logger

        # Read dataset example
        with dataset_op_logger(path, "read") as logger:
            df = read(path, ...)
            logger.set(data=df)

        # Write dataset example
        df = DataFrame(data)
        with dataset_op_logger(path, "write") as logger:
            write(df, ...)
            logger.set(data=df)


        with dataset_op_logger("location://path/to/value.csv", "read"):
            value = read_from()
            logger.set(data=value)
            # Read is successful

        calculate_alpha()


        with dataset_op_logger("location://path/to/value.csv", "read") as logger:
            value = read_from()
            logger.set(data=value)
            # Read is successful
            calculate_alpha()
            # If calculate_alpha raises an exception, a failed read operation is reported to Databand.


        @task()
        def prepare_data():
            log_dataset_op(
                "/path/to/value.csv",
                DbndDatasetOperationType.read,
                data=pandas_data_frame,
                with_preview=True,
                with_schema=True,
            )

        import databricks.koalas as ks
        from targets.values import DataFrameValueType, register_value_type

        class KoalasValueType(DataFrameValueType):
            type = ks.DataFrame
            type_str = "KoalasDataFrame"

        register_value_type(KoalasValueType())
        #### DOC END
        prepare_data.dbnd_run()
        """



from time import time

from mxnet import autograd, gluon, init
from mxnet.gluon import nn
from mxnet.gluon.data.dataset import ArrayDataset
from mxnet.gluon.data.vision import transforms

from dbnd import data, namespace, output, parameter
from dbnd.tasks import PipelineTask
from dbnd_examples.orchestration.tool_mxnet import (
    DownloadFile,
    MxNetGluonTask,
    fashion_data,
    read_minst,
)


namespace("fashion")


def accuracy(output, label):
    return (output.argmax(axis=1) == label.astype("float32")).mean().asscalar()


def build_lenet_classifier():
    net = nn.Sequential()
    net.add(
        nn.Conv2D(channels=6, kernel_size=5, activation="relu"),
        nn.MaxPool2D(pool_size=2, strides=2),
        nn.Conv2D(channels=16, kernel_size=3, activation="relu"),
        nn.MaxPool2D(pool_size=2, strides=2),
        nn.Flatten(),
        nn.Dense(120, activation="relu"),
        nn.Dense(84, activation="relu"),
        nn.Dense(10),
    )
    net.initialize(init=init.Xavier())
    return net


def forward_backward_path(
    network, loss_function, acc_function, train_data, trainer, batch_size
):
    train_loss, train_acc = 0.0, 0.0
    for data, label in train_data:
        # forward + backward
        with autograd.record():
            output = network(data)
            loss = loss_function(output, label)
        loss.backward()
        # update parameters
        trainer.step(batch_size)
        # calculate training metrics
        train_loss += loss.mean().asscalar()
        train_acc += acc_function(output, label)
    return train_acc, train_loss


class TrainModel(MxNetGluonTask):
    learning_rate = parameter.value(0.1)
    optimizer = parameter.value("sgd")
    epoch_num = parameter.value(10)
    batch_size = 256

    training_data = data
    training_labels = data

    test_data = data
    test_labels = data

    model = output

    def _get_data_loader(self, data, labels):
        # prepare data
        data, label = read_minst(str(data), str(labels), gluon_format=True)
        transform_data = transforms.Compose(
            [transforms.ToTensor(), transforms.Normalize(0.13, 0.31)]
        )
        data_set = ArrayDataset(data, label).transform_first(transform_data)

        return gluon.data.DataLoader(data_set, batch_size=self.batch_size, shuffle=True)

    def get_trainer(self, network):
        return gluon.Trainer(
            network.collect_params(),
            self.optimizer,
            {"learning_rate": self.learning_rate},
        )

    def main(self, ctx):
        # prepare data
        train_dataloader = self._get_data_loader(
            self.training_data, self.training_labels
        )
        test_dataloader = self._get_data_loader(self.test_data, self.test_labels)

        network = build_lenet_classifier()
        trainer = self.get_trainer(network)

        # train
        for epoch in range(self.epoch_num):
            train_loss, train_acc, valid_acc = 0.0, 0.0, 0.0
            tic = time()
            train_acc, train_loss = forward_backward_path(
                network,
                gluon.loss.SoftmaxCrossEntropyLoss(),
                accuracy,
                train_dataloader,
                trainer,
                self.batch_size,
            )

            # calculate validation accuracy
            for data, label in test_dataloader:
                valid_acc += accuracy(network(data), label)

            self.log_metric(
                "epoch",
                (
                    epoch,
                    train_loss / len(train_dataloader),
                    train_acc / len(train_dataloader),
                    valid_acc / len(test_dataloader),
                    time() - tic,
                ),
            )

        network.save_parameters(self.model.path)


class PredictClothingType(PipelineTask):
    model = output

    def band(self):
        train_data = DownloadFile.from_web(fashion_data.train_data).data
        train_labels = DownloadFile.from_web(fashion_data.train_label).data
        test_data = DownloadFile.from_web(fashion_data.test_data).data
        test_labels = DownloadFile.from_web(fashion_data.test_label).data

        model_task = TrainModel(
            training_data=train_data,
            training_labels=train_labels,
            test_data=test_data,
            test_labels=test_labels,
        )
        self.model = model_task.model



"""
class TestDocAirflowAndSnowflake:
    def test_doc(self):
        #### DOC START
        import airflow

        from airflow.contrib.hooks.snowflake_hook import SnowflakeHook
        from airflow.models import DAG, BaseOperator

        args = {"start_date": airflow.utils.dates.days_ago(2), "owner": "databand"}

        snowflake_conn_id = "test_snowflake_conn"
        database = "SNOWFLAKE_SAMPLE_DATA"
        schema = "TPCDS_SF100TCL"
        table = "CUSTOMER"
        select_query = f"select * from {database}.{schema}.{table} limit 1000"

        example_dags = [
            DAG(
                dag_id="example_dag_{}".format(i),
                default_args=args,
                schedule_interval="0 1 * * *",  # Daily
            )
            for i in range(1, 4)
        ]
        for dag in example_dags:
            globals()[dag.dag_id] = dag



        from dbnd_snowflake import log_snowflake_resource_usage, log_snowflake_table
        from dbnd_snowflake.airflow_hooks import DbndSnowflakeHook

        class CalculateAlpha(BaseOperator):
            def execute(self, context):
                hook = DbndSnowflakeHook(snowflake_conn_id=snowflake_conn_id)
                session_id, query_ids = hook.run(sql=select_query)
                connection_string = hook.get_uri()

                log_snowflake_table(
                    table_name=table,
                    connection_string=connection_string,
                    database=database,
                    schema=schema,
                    key=f"calculate_alpha.{table}",
                    with_preview=False,
                    raise_on_error=False,
                )

                log_snowflake_resource_usage(
                    database=database,
                    key=f"calculate_alpha.res.{session_id}{query_ids[0]}",
                    connection_string=connection_string,
                    query_ids=query_ids,
                    session_id=int(session_id),
                    raise_on_error=True,
                )

        with example_dags[0]:
            calculate_alpha = CalculateAlpha(task_id="calculate_alpha")



        from dbnd_snowflake import log_snowflake_resource_usage, log_snowflake_table
        from dbnd_snowflake.airflow_hooks import snowflake_run

        class CalculateBeta(BaseOperator):
            def execute(self, context):
                hook = SnowflakeHook(snowflake_conn_id=snowflake_conn_id)
                session_id, query_ids = snowflake_run(hook, select_query)
                connection_string = hook.get_uri()

                log_snowflake_table(
                    table_name=table,
                    connection_string=connection_string,
                    database=database,
                    schema=schema,
                    key=f"calculate_beta.{table}",
                    with_preview=False,
                    raise_on_error=False,
                )

                log_snowflake_resource_usage(
                    database=database,
                    key=f"calculate_beta.{session_id}{query_ids[0]}",
                    connection_string=connection_string,
                    query_ids=query_ids,
                    session_id=int(session_id),
                    raise_on_error=True,
                )

        with example_dags[1]:
            calculate_beta = CalculateBeta(task_id="calculate_beta")



        from dbnd_snowflake.airflow_hooks import snowflake_run
        from dbnd_snowflake.airflow_operators import (
            LogSnowflakeResourceOperator,
            LogSnowflakeTableOperator,
        )

        with example_dags[2]:
            calculate_coefficient = CalculateCoefficient(task_id="calculate_coefficient")
            log_snowflake_table_task = LogSnowflakeTableOperator(
                table=table,
                snowflake_conn_id=snowflake_conn_id,
                warehouse=None,
                database=database,
                schema=schema,
                task_id="log_snowflake_table_task",
                key=f"snowflake_table_{table}",
            )

            log_snowflake_resources_task = LogSnowflakeResourceOperator(
                session_id="{{ti.xcom_pull(key='session_id') }}",
                query_id="{{ti.xcom_pull(key='query_id')[0] }}",
                snowflake_conn_id=snowflake_conn_id,
                warehouse=None,
                database=database,
                schema=schema,
                task_id="log_snowflake_resources_task",
                key="snowflake_query_resources",
            )
            calculate_coefficient >> log_snowflake_table_task >> log_snowflake_resources_task
        #### DOC END
"""



import logging

from dbnd.testing.helpers_pytest import assert_run_task
from dbnd_examples.orchestration.tool_sklearn.sklearn_example import linear_reg_pipeline


logger = logging.getLogger(__name__)
logging.basicConfig(level=logging.INFO)


def test_sklearn_tracking():
    task = assert_run_task(linear_reg_pipeline.task())
    assert task is not None
    assert task.result is not None



"""
class TestDocTrackingRedshift:
    def test_doc(self):
        #### DOC START
        import psycopg2
        from dbnd_redshift.dbnd_redshift import get_redshift_query_id, log_redshift_resource_usage

        def establish_connection():
            '''establish and return a Redshift connection'''
            redshift_conn = psycopg2.connect(
                dbname='database',
                host='cluster-name.acc-name.region.redshift.amazonaws.com',
                port='5439', user='user', password='password'
            )
            return redshift_conn

        def execute_and_log(redshift_connection, query):
            '''execute a query and log Redshift resource usage'''
            cursor = redshift_connection.cursor()
            cursor.execute(query)

            query_id = get_redshift_query_id(query, connection=redshift_connection)
            log_redshift_resource_usage(query_id, connection=redshift_connection)

        if __name__ == "__main__":
            redshift_connection = establish_connection()

            query = "SELECT * FROM transaction_data WHERE transaction_amt >= 5000;"
            execute_and_log(redshift_connection, query)
            print("works")
        #### DOC END
        """



import os

from datetime import timedelta

import airflow

from airflow import DAG
from airflow.contrib.hooks.snowflake_hook import SnowflakeHook
from airflow.operators.python_operator import PythonOperator

from dbnd_snowflake import snowflake_query_tracker
from dbnd_snowflake.airflow_operators import LogSnowflakeTableOperator


user = os.environ.get("SNOWFLAKE_USER")
account = os.environ.get("SNOWFLAKE_ACCOUNT")
database = "SNOWFLAKE_SAMPLE_DATA"
schema = "TPCDS_SF100TCL"
table = "CUSTOMER"
SNOWFLAKE_CONNECTION_ID = "test_snowflake_conn"

args = {"start_date": airflow.utils.dates.days_ago(2), "owner": "databand"}

select_query = f'select c_birth_year from "{database}"."{schema}"."{table}" where C_SALUTATION = "Dr." limit 1000'
# Update query is expected to fail
update_query = f'UPDATE "{database}"."{schema}"."{table}" SET C_CUSTOMER_ID=C_CUSTOMER_ID WHERE C_SALUTATION = "Dr."'


def process_records(records):
    return min(r[0] for r in records)


def process_customers(**kwargs):
    snowflake_hook = SnowflakeHook(snowflake_conn_id=SNOWFLAKE_CONNECTION_ID)
    customers = snowflake_hook.get_records(select_query)

    # Process records
    process_records(customers)


def process_customers_with_monitoring(**kwargs):
    snowflake_hook = SnowflakeHook(snowflake_conn_id=SNOWFLAKE_CONNECTION_ID)
    with snowflake_query_tracker(log_tables=False, database=database):
        customers = snowflake_hook.get_records(select_query)
    # Process records - Same code
    process_records(customers)


def update_customers(**kwargs):
    snowflake_hook = SnowflakeHook(snowflake_conn_id=SNOWFLAKE_CONNECTION_ID)
    snowflake_hook.run(update_query)


def update_customers_with_monitoring(**kwargs):
    snowflake_hook = SnowflakeHook(snowflake_conn_id=SNOWFLAKE_CONNECTION_ID)
    with snowflake_query_tracker(log_tables=False, database=database):
        snowflake_hook.run(update_query)


dag_no_dbnd = DAG(
    dag_id="snowflake_no_monitoring_dag",
    default_args=args,
    schedule_interval=timedelta(hours=6),
)

with dag_no_dbnd:
    get_customers = PythonOperator(
        task_id="get_customers_no_dbnd", python_callable=process_customers
    )
    update_db = PythonOperator(
        task_id="update_customers_no_dbnd", python_callable=update_customers
    )

    get_customers >> update_db

dag_with_dbnd = DAG(
    dag_id="snowflake_with_monitoring_dag",
    default_args=args,
    schedule_interval=timedelta(hours=6),
)

with dag_with_dbnd:
    get_customers_dbnd = PythonOperator(
        task_id="get_customers_with_dbnd",
        python_callable=process_customers_with_monitoring,
    )
    update_db_dbnd = PythonOperator(
        task_id="update_customers_no_dbnd",
        python_callable=update_customers_with_monitoring,
    )
    log_snowflake_table_task = LogSnowflakeTableOperator(
        table=table,
        snowflake_conn_id=SNOWFLAKE_CONNECTION_ID,
        warehouse=None,
        database=database,
        schema=schema,
        account=account,
        task_id="log_snowflake_table_task",
        key=f"snowflake_table_{table}",
        with_preview=True,
        with_schema=True,
        raise_on_error=True,
    )

    get_customers_dbnd >> update_db_dbnd >> log_snowflake_table_task



from dbnd import parameter
from dbnd._core.settings.tracking_config import ValueTrackingLevel


class TestDocPerformantLogging:
    def test_doc(self):
        #### DOC START
        value_reporting_strategy = parameter(
            default=ValueTrackingLevel.SMART,
            description="Multiple strategies with different limitations on potentially expensive calculation for value_meta."
            "ALL => no limitations."
            "SMART (default) => restrictions on lazy evaluation types."
            "NONE => limit everything.",
        ).enum(ValueTrackingLevel)
        #### DOC END



import logging

from datetime import datetime, timedelta
from typing import Tuple

import airflow

from airflow import DAG
from airflow.utils.dates import days_ago

from dbnd import parameter, pipeline, task


default_args = {
    "owner": "airflow",
    "depends_on_past": False,
    "start_date": days_ago(2),
    "retries": 1,
    "retry_delay": timedelta(minutes=5),
}


class TestDocDecoratedDags:
    def test_code_example(self):
        #### DOC START
        # support airflow 1.10.0
        from airflow.operators.python_operator import PythonOperator

        if airflow.version.version == "1.10.0":

            class PythonOperator_airflow_1_10_0(PythonOperator):
                template_fields = ("templates_dict", "op_kwargs")

            PythonOperator = PythonOperator_airflow_1_10_0

        @task
        def calculate_alpha(alpha: int):
            logging.info("I am running")
            return alpha

        @task
        def prepare_data(data: str) -> Tuple[str, str]:
            return data, data

        def read_and_write(input_path, output_path):
            logging.error("I am running")
            input_value = open(input_path, "r").read()
            with open(output_path, "w") as fp:
                fp.write(input_value)
                fp.write("\n\n")
                fp.write(str(datetime.now().strftime("%Y-%m-%dT%H:%M:%S")))
            return "success"

        with DAG(dag_id="dbnd_operators", default_args=default_args) as dag_operators:
            # t1, t2 and t3 are examples of tasks created by instantiating operators
            t1 = calculate_alpha(2)
            t2, t3 = prepare_data(t1)
            tp = PythonOperator(
                task_id="some_python_function",
                python_callable=read_and_write,
                op_kwargs={"input_path": t3, "output_path": "/tmp/output.txt"},
            )
            tp.set_upstream(t3.op)

            t1_op = t1.op

            airflow_op_kwargs = {"priority_weight": 50}
            # Define DAG context
            with DAG(
                dag_id="dbnd_operators", default_args=default_args
            ) as dag_operators:
                t1 = calculate_alpha(2, task_airflow_op_kwargs=airflow_op_kwargs)
        #### DOC END
        assert t1_op

    def test_prepare_data_jinja_templating(self):
        #### DOC START
        @pipeline
        def current_date(p_date=None):
            return p_date

        with DAG(dag_id=f"current_date_dag", default_args=default_args) as dag:
            current_date(p_date="{{ ts }}")
        #### DOC END

    def test_prepare_data_no_jinja_templating(self):
        #### DOC START
        @pipeline
        def current_date(p_date=parameter[str].disable_jinja_templating):
            return p_date

        with DAG(dag_id=f"current_date_dag", default_args=default_args) as dag:
            current_date(p_date="{{ ts }}")
        #### DOC END



import os

import airflow

from airflow import DAG
from airflow.hooks.postgres_hook import PostgresHook
from airflow.operators.python_operator import PythonOperator
from numpy import issubdtype, number

from dbnd import log_dataframe, log_metric


REDSHIFT_CONNECTION_ID = os.getenv("REDSHIFT_CONN", default="redshift_conn")
REDSHIFT_TABLE = os.getenv(
    "REDSHIFT_MONITOR_TARGET_TABLE", default="information_schema.columns"
)
REDSHIFT_TABLE_MONITOR_SCHEDULE = os.getenv(
    "REDSHIFT_TABLE_MONITOR_SCHEDULE", default="@daily"
)
REDSHIFT_MONITOR_TABLE_LIMIT = os.getenv("REDSHIFT_MONITOR_TABLE_LIMIT", default=1000)

# query:
SELECT_DATA = "SELECT * FROM {} LIMIT %s;".format(REDSHIFT_TABLE)

DEFAULT_ARGS = {
    "owner": "databand",
    "start_date": airflow.utils.dates.days_ago(0),
    "provide_context": False,
}


def monitor_redshift_table(**op_kwarg):
    """Redshift table monitor collects the following metrics:
    - record count
    - duplicate records
    - Null/NaN record counts in each column
    - mean, median, min, max, std of each numeric column
    """

    hook = PostgresHook(REDSHIFT_CONNECTION_ID)
    data = hook.get_pandas_df(SELECT_DATA, parameters=[REDSHIFT_MONITOR_TABLE_LIMIT])

    log_dataframe(
        "{}".format(REDSHIFT_TABLE),
        data,
        with_histograms=True,
        with_stats=True,
        with_schema=True,
    )

    log_metric("record count", data.shape[0])
    log_metric("Duplicate records", data.shape[0] - data.drop_duplicates().shape[0])
    for column in data.columns:
        log_metric(
            "{} null record count".format(column), int(data[column].isna().sum())
        )

        if issubdtype(data[column].dtype, number):
            log_metric("{} mean".format(column), round(data[column].mean(), 2))
            log_metric("{} median".format(column), data[column].median())
            log_metric("{} min".format(column), data[column].min())
            log_metric("{} max".format(column), data[column].max())
            log_metric("{} std".format(column), round(data[column].std(), 2))


with DAG(
    dag_id="dbnd_redshift_{}_monitor".format(REDSHIFT_TABLE),
    schedule_interval="{}".format(REDSHIFT_TABLE_MONITOR_SCHEDULE),
    default_args=DEFAULT_ARGS,
) as dbnd_template_dag:

    redshift_monitor = PythonOperator(
        task_id="monitor_redshift_table", python_callable=monitor_redshift_table
    )



import time

from pprint import pprint

import airflow
import pandas as pd

from airflow.models import DAG
from airflow.operators.python_operator import PythonOperator

from dbnd import log_dataframe, log_metric, task
from dbnd._vendor.pendulum import utcnow


args = {"owner": "airflow", "start_date": airflow.utils.dates.days_ago(2)}

dag = DAG(dag_id="embed_dbnd", default_args=args, schedule_interval=None)


def print_context(ds, **kwargs):
    """Print the Airflow context and ds variable from the context."""
    pprint(kwargs)
    log_metric("asd", 1.1)
    print(ds)
    return "Whatever you return gets printed in the logs"


run_this = PythonOperator(
    task_id="print_the_context",
    provide_context=True,
    python_callable=print_context,
    dag=dag,
)
# [END howto_operator_python]


def stub(stage, test_df=None):
    # type: (str, pd.DataFrame) -> None
    # if random.randint(0, 1):
    #     raise Exception("brrrr")

    log_metric(stage, utcnow())
    log_dataframe("df_" + stage, test_df)


# [START howto_operator_python_kwargs]
def my_sleeping_function(random_base):
    """This is a function that will run within the DAG execution"""
    df = pd.DataFrame(
        {
            "num_legs": [2, 4, 8, 0],
            "num_wings": [2, 0, 0, 0],
            "num_specimen_seen": [10, 2, 1, 8],
        },
        index=["falcon", "dog", "spider", "fish"],
    )
    stub("before", df.sample(n=3, random_state=1))
    time.sleep(random_base)
    stub("after", df.sample(frac=0.5, replace=True, random_state=1))


# Generate 5 sleeping tasks, sleeping from 0.0 to 0.4 seconds respectively
for i in range(5):
    task = PythonOperator(
        task_id="sleep_for_" + str(i),
        python_callable=my_sleeping_function,
        op_kwargs={"random_base": float(i) / 10},
        dag=dag,
    )

    run_this >> task



import os

import airflow

from airflow import DAG
from airflow.hooks.postgres_hook import PostgresHook
from airflow.operators.python_operator import PythonOperator
from numpy import mean, median
from pandas import DataFrame

from dbnd import log_metric


# get environment variables:
REDSHIFT_CONNECTION_ID = os.getenv("REDSHIFT_CONN", default="redshift_conn")
REDSHIFT_CLUSTER_NAME = os.getenv("REDSHIFT_CLUSTER_NAME", default="redshift_cluster")
REDSHIFT_CLUSTER_MONITOR_SCHEDULE = os.getenv(
    "REDSHIFT_CLUSTER_MONITOR_SCHEDULE", default="@daily"
)
REDSHIFT_SCHEMA = os.getenv("REDSHIFT_SCHEMA", default="public")

# Queries:
COUNT_TABLES = """
select count(*)
from information_schema.tables
where table_schema = %s;
"""

COUNT_TABLE_ROWS = """
select svv.table_name,
       table_info.tbl_rows as row_count
from svv_tables svv
join svv_table_info table_info
          on svv.table_name = table_info.table
where svv.table_schema = %s
    and svv.table_type = 'BASE TABLE';
"""

DESCRIBE_TABLES = """
select * from pg_table_def where schemaname = %s;
"""

DISK_USAGE = """
select
  round(sum(capacity),2)/1024 as capacity_gb,
  round (sum(used),2)/1024 as used_gb,
  round((sum(capacity) - sum(used)),2)/1024 as free_gb
from
  stv_partitions where part_begin=0;
"""

DEFAULT_ARGS = {
    "owner": "databand",
    "start_date": airflow.utils.dates.days_ago(0),
    "provide_context": False,
}


def monitor_redshift(**op_kwarg):
    """Redshift database monitor collects the following metrics:
    - Number of tables in database
    - Shape of each table in the database
    - Min, max, mean, median number of rows across all tables,
    - Min, max, mean, median number of columns across all tables,
    - Total number of rows and columns
    - Largest tables by row and column
    - Disk capacity, Free space on disk, Used space on disk (in GB)
    - Disk percent usage
    """
    hook = PostgresHook(REDSHIFT_CONNECTION_ID)
    num_redshift_tables = hook.get_first(COUNT_TABLES, parameters=[REDSHIFT_SCHEMA])[0]
    log_metric("Cluster table count", num_redshift_tables)

    table_row_counts = hook.get_records(COUNT_TABLE_ROWS, parameters=[REDSHIFT_SCHEMA])
    num_rows_per_table = {}
    for tablename, row_count in table_row_counts:
        num_rows_per_table[tablename] = int(round(row_count))

    row_counts = list(num_rows_per_table.values())
    log_metric("Max table row count", max(row_counts))
    log_metric("Min table row count", min(row_counts))
    log_metric("Mean table row count", round(mean(row_counts), 2))
    log_metric("Median table row count", median(row_counts))

    tables = hook.get_pandas_df(DESCRIBE_TABLES, parameters=[REDSHIFT_SCHEMA])
    table_shapes = DataFrame()
    table_shapes["columns"] = tables.groupby("tablename").nunique("column")["column"]
    table_shapes["tablename"] = tables["tablename"].unique()
    table_shapes["rows"] = (
        table_shapes["tablename"].map(num_rows_per_table).fillna(0).astype(int)
    )

    for _, row in table_shapes.iterrows():
        log_metric("{} shape".format(row["tablename"]), (row["columns"], row["rows"]))

    log_metric("Max table column count", table_shapes["columns"].max())
    log_metric("Min table column count", table_shapes["columns"].max())
    log_metric("Mean table column count", round(table_shapes["columns"].mean(), 2))
    log_metric("Median table column count", table_shapes["columns"].median())

    log_metric("Total columns", table_shapes["columns"].sum())
    log_metric("Total rows", table_shapes["rows"].sum())

    max_row_table = table_shapes[table_shapes["rows"] == table_shapes["rows"].max()]
    max_col_table = table_shapes[
        table_shapes["columns"] == table_shapes["columns"].max()
    ]
    log_metric("Largest table (by row count)", max_row_table["tablename"][0])
    log_metric("Largest table (by col count)", max_col_table["tablename"][0])

    disk_stats = hook.get_records(DISK_USAGE).pop()
    disk_capacity, disk_used, disk_free = disk_stats
    log_metric("Disk capacity (GB)", disk_capacity)
    log_metric("Disk used (GB)", disk_used)
    log_metric("Disk free (GB)", disk_free)
    log_metric("Percent Disk usage", round((disk_used / disk_capacity) * 100, 2))


with DAG(
    dag_id="dbnd_{}_monitor".format(REDSHIFT_CLUSTER_NAME),
    schedule_interval="{}".format(REDSHIFT_CLUSTER_MONITOR_SCHEDULE),
    default_args=DEFAULT_ARGS,
) as dbnd_template_dag:

    redshift_monitor = PythonOperator(
        task_id="monitor_redshift_cluster", python_callable=monitor_redshift
    )
